#!/usr/bin/python
# All Rights Reserved. Copyright (c) 2018, Hitachi, Ltd.       #
# Licensed Material of Hitachi, Ltd.                           #
# Reproduction, use, modification or disclosure otherwise than #
# permitted in the License Agreement is strictly prohibited.   #
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"

import os
import sys
import errno
import ConfigParser

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
from CommonLogger import LogLevel
import FailoverConfig

def_config_file = "conf/foscripts.conf"

# Special logger definition.
# Just output to stdout as it is expected no logger has been
# configured yet at this stage.
def log_error(msg):
    print msg
log_info = log_warning = log_error

class BaseError(Exception):
    pass

class FormatError(BaseError):
    pass

class ConfigurationError(BaseError):
    pass

class ParameterError(BaseError):
    pass

class Config(object):
    def __init__(self, config_file=None):
        # Configurable parameters
        self.config_file = None
        self.hdi_conf_dir = None
        self.sys_conf_dir = None
        self.fo_conf_dir = None
        self.log_dir = None
        self.retry_count = None
        self.retry_interval = None
        self.timeout = None
        self.default_policy_string = None
        self.log_level = None
        self.log_size = None
        self.log_gen = None
        self.print_msg = None
        self.send_alert = None
        self.sleep_for_wait = None

        # Fixed parameters
        self.ssh_user = "nasroot"
        self.rest_user = "admin"
        self.ssh_key_name = "sshloginOpenSSH_%s"
        self.ssh_default_key_name = "sshloginOpenSSH_default"
        self.ssh_key_public = None
        self.ssh_key_private = None
        self.ssh_default_key_public = None
        self.ssh_default_key_private = None
        self.hdi_auth_path = None
        self.rest_passMD5 = None
        self.common_log_file = "hdi_failover_scripts.log"

        # Given parameters
        # FailoverConfig.PairConfig to determine SSH key to be used.
        self.pairconf = None

        # Load config file
        self.config_obj = ConfigParser.SafeConfigParser()
        if config_file is None:
            self.config_file = os.path.join(BASEDIR, def_config_file)
        else:
            self.config_file = config_file

        try:
            if self.config_obj.read(self.config_file) == []:
                raise FormatError("Cannot read config file %s" % self.config_file)
            dir_section = "Directory"
            self.hdi_conf_dir = self._get_config_string(dir_section, "HDIConfDir")
            self.fo_conf_dir = self._get_config_string(dir_section, "FailoverConfDir")
            self.sys_conf_dir = self._get_config_string(dir_section, "SysConfDir")
            self.log_dir = self._get_config_string(dir_section, "LogDir")

            comm_section = "Communication"
            self.retry_count = self._get_config_integer(comm_section, "RetryCount", allow_zero=True)
            self.retry_interval = self._get_config_integer(comm_section, "RetryInterval")
            self.timeout = self._get_config_integer(comm_section, "Timeout", allow_zero=True)

            func_section = "Function"
            self.default_policy_string = self._get_config_string(func_section, "DefaultStoragePolicy")
            # Check storage policy configuration format
            try:
                policy = FailoverConfig.StoragePolicy("test", policy_string=self.default_policy_string)
            except FailoverConfig.FormatError, e:
                raise ValueError("Invalid format for DefaultStoragePolicy: %s" % self.default_policy_string)
            if not policy.policy in (FailoverConfig.StoragePolicy.AUTOTOTAL, FailoverConfig.StoragePolicy.AUTOUSAGE, FailoverConfig.StoragePolicy.AUTOSPEC):
                raise ValueError("Invalid storage policy as a default: %s" % self.default_policy_string)
            self.sleep_for_wait = self._get_config_integer(func_section, "SleepForWait")

            log_section = "Log"
            log_level_str = self._get_config_string(log_section, "LogLevel")
            if log_level_str == "DEBUG":
                self.log_level = LogLevel.DEBUG
            elif log_level_str == "INFO":
                self.log_level = LogLevel.INFO
            elif log_level_str == "WARNING":
                self.log_level = LogLevel.WARNING
            elif log_level_str == "ERROR":
                self.log_level = LogLevel.ERROR
            else:
                self.log_level = LogLevel.INFO
            self.log_size = self._get_config_integer(log_section, "Size")
            self.log_gen = self._get_config_integer(log_section, "Generation")
            self.print_msg = self._get_config_yesno(log_section, "PrintMessage")
            self.send_alert = self._get_config_yesno(log_section, "SendAlert")
        except ValueError, e:
            log_error("Invalid parameter was specified in the configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise FormatError("An invalid information is recorded in config file. file=%s, error=%s" % (self.config_file, str(e)))
        except (ConfigParser.MissingSectionHeaderError, ConfigParser.NoSectionError, ConfigParser.NoOptionError), e:
            log_error("Required information is missing in config file. file=%s, error=%s." % (self.config_file, str(e)))
            raise FormatError("An invalid information is recorded in config file. file=%s, error=%s." % (self.config_file, str(e)))
        except ConfigParser.Error, e:
            log_error("Failed to parse config file. file=%s, error=%s" % (self.config_file, str(e)))
            raise FormatError("An invalid information is recorded in config file. file=%s, error=%s" % (self.config_file, str(e)))

        # Initialize as absolute path
        if not os.path.isabs(self.hdi_conf_dir):
            self.hdi_conf_dir = os.path.abspath(os.path.join(BASEDIR, self.hdi_conf_dir))
        if not os.path.isabs(self.fo_conf_dir):
            self.fo_conf_dir = os.path.abspath(os.path.join(BASEDIR, self.fo_conf_dir))
        if not os.path.isabs(self.sys_conf_dir):
            self.sys_conf_dir = os.path.abspath(os.path.join(BASEDIR, self.sys_conf_dir))
        if not os.path.isabs(self.log_dir):
            self.log_dir = os.path.abspath(os.path.join(BASEDIR, self.log_dir))

        # Create directories if not exist
        try:
            if not os.path.isdir(self.hdi_conf_dir):
                os.mkdir(self.hdi_conf_dir)
            if not os.path.isdir(self.fo_conf_dir):
                os.mkdir(self.fo_conf_dir)
            if not os.path.isdir(self.sys_conf_dir):
                os.mkdir(self.sys_conf_dir)
            if not os.path.isdir(self.log_dir):
                os.mkdir(self.log_dir)
        except EnvironmentError, e:
            log_error("Failed to create a system directory. err=%s" % str(e))
            raise

        # Set fixed parameters
        ssh_key_path = os.path.join(self.sys_conf_dir, self.ssh_key_name)
        self.ssh_key_public = ssh_key_path + ".puk"
        self.ssh_key_private = ssh_key_path + ".ppk"
        ssh_key_path = os.path.join(self.sys_conf_dir, self.ssh_default_key_name)
        self.ssh_default_key_public = ssh_key_path + ".puk"
        self.ssh_default_key_private = ssh_key_path + ".ppk"
        self.hdi_auth_path = os.path.join(self.sys_conf_dir, "hdi_dr_auth")
        try:
            with open(self.hdi_auth_path, 'r') as f:
                auth_token = f.readline().strip()
                notused, passMD5 = auth_token.split(':')
                self.rest_passMD5 = passMD5
        except EnvironmentError, e:
            if e.errno == errno.ENOENT:
                log_error("API authentication configuration does not exist.")
                raise ConfigurationError("API authentication configuration does not exist.")
            else:
                log_error("Failed to open/read API authentication configuration file. error=%s" % str(e))
                raise

    def get_ssh_private_key(self, ipaddr=None):
        """ Returns an SSH private key path used for connecting a node.

        This function determines which SSH key (default or unique) should be
        used to connect a node based on the PairConfig information.

        If no PairConfig is set, returns default SSH key.

        If an IP address is specified, checks if the node (represented by
        the IP address) already has Unique key or not. If it has, returns
        Unique key. If it doesn't, returns default key.

        If PairConfig is set AND no IP address is specified, returns Unique key.

        Args:
            ipaddr (str): An IP address of a node to connect.
        Returns:
            str: An SSH private key path to be used to connect the node.
        """
        if self.pairconf is None:
            private_key = self.ssh_default_key_private
        else:
            if ipaddr is None:
                private_key = self.ssh_key_private % self.pairconf.confgroup
            elif self.pairconf.is_ssh_setup(ipaddr):
                private_key = self.ssh_key_private % self.pairconf.confgroup
            else:
                private_key = self.ssh_default_key_private

        return private_key

    def get_ssh_public_key(self, confgroup=None):
        if self.pairconf is None:
            if confgroup is None:
                public_key = self.ssh_default_key_public
            else:
                public_key = self.ssh_key_public % confgroup
        else:
            public_key = self.ssh_key_public % self.pairconf.confgroup

        return public_key

    def _get_config_string(self, section, item):
        return self.config_obj.get(section, item).strip('"')

    def _get_config_integer(self, section, item, allow_zero=False):
        value = self.config_obj.get(section, item)
        if not value.isdigit():
            log_error("%s is not a number. %s = %s" % (item, item, value))
            raise ConfigurationError("An invalid parameter was found. file=%s" % self.config_file)
        value = int(value)
        if (allow_zero and value < 0) or (not allow_zero and value <= 0):
            log_error("%s cannot be 0 or negative. %s = %s" % (item, item, value))
            raise ConfigurationError("An invalid parameter was found. file=%s" % self.config_file)

        return value

    def _get_config_yesno(self, section, item):
        value = self.config_obj.get(section, item)
        if value.lower() != "yes" and value.lower() != "no":
            log_error("Invalid config parameter: %s = %s" % (item, value))
            raise ConfigurationError("An invalid config parameter was found. file=%s" % self.config_file)

        return True if value.lower() == "yes" else False

class SSHConfig(object):
    SSH_USER = 'nasroot'
    BASE_KEY_NAME = 'sshloginOpenSSH'
    DEFAULT_KEY_NAME = BASE_KEY_NAME + '_default'

    def __init__(self, config_dir, pairconf=None, confgroup=None):
        self.config_dir = config_dir
        self.pairconf = pairconf
        self.confgroup = confgroup
        self.public_key = None
        self.private_key = None
        self.ssh_user = self.SSH_USER   # Fixed parameter

    def get_public_key(self, is_default=False):
        key = None
        if is_default:
            key = self._make_default_public_key_path()
        else:
            if self.confgroup is None:
                raise ParameterError("Configuration group must be specified to get Unique key name.")
            key = self._make_unique_public_key_path()

        return key

    def get_private_key(self, ipaddr=None, is_default=False):
        key = None
        if is_default:
            key = self._make_default_private_key_path()
        else:
            if self.confgroup is None:
                raise ParameterError("Configuration group must be specified to get Unique key name.")

            if ipaddr is None:
                # Always returns Unique key name is no IP address is specified.
                key = self._make_unique_private_key_path()
            elif self.pairconf is None:
                # Need PairConfig to decide if the provided IP address node has Unique key or not.
                raise ParameterError("PairConfig instance must be specified to get Unique key of a specified node.")
            else:
                # Both PairConfig and IP address are provided.
                # Decide an appropriate key (default or unique) for the node.
                if self.pairconf.is_ssh_setup(ipaddr):
                    key = self._make_unique_private_key_path()
                else:
                    key = self._make_default_private_key_path()

        return key

    def _make_default_public_key_path(self):
        return os.path.join(self.config_dir, self.DEFAULT_KEY_NAME + '.puk')

    def _make_default_private_key_path(self):
        return os.path.join(self.config_dir, self.DEFAULT_KEY_NAME + '.ppk')

    def _make_unique_public_key_path(self):
        return os.path.join(self.config_dir, self.BASE_KEY_NAME + '_%s' % self.confgroup + '.puk')

    def _make_unique_private_key_path(self):
        return os.path.join(self.config_dir, self.BASE_KEY_NAME + '_%s' % self.confgroup + '.ppk')
