#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "execute_filesystem_failover"

import sys
import os
import errno
import fcntl
import getopt
import time

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_error

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import Alert
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import failover_common
import lock_common
import backup_sync_common
import failover_filesystem

def usage():
    print "Usage:"
    print "    sudo maintexec %s [--dry-run] [--try-out] [--error]" % PROGRAM
    print "    sudo maintexec %s -h|--help" % PROGRAM

def execute_filesystem_failover(config, is_error=False, rehearsal=None):
    """ Execute registered failover jobs.

    This function executes failover jobs one by one in sequential number order.
    Finished job is deleted.
    This function executes failover jobs until all jobs are deleted. This means,
    if a new job is added while this function executing a failover, the new job
    will also be executed later in this function.

    Configuration object (config) is required to get pair/storage policy configuration
    if no storage policy is provided in a failover job.

    Args:
        config (Config.Config): Configuration of the scripts.
        is_error (bool): True if you want to execute failover jobs that finished
            with an error previously.
        rehearsal (RehearsalMode): Rehearsal mode if this is rehearsal failover.

    Returns:
        bool: True if an error occurred while executing an failover job. Otherwise False.
    """

    # Get lock for job execution
    print_status("Checking parallel execution ...")
    execute_lock_path = lock_common.get_execute_lock_file()
    locker = open(execute_lock_path, 'w')
    while True:
        try:
            # This lock will be released when the process finishes
            fcntl.flock(locker, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except EnvironmentError, e:
            if e.errno == errno.EAGAIN:
                # The lock is already acquired by another process. Wait for 60 seconds.
                print_status("File system synchronization or failover execution is running on this node. Waiting %s seconds and then will retry ..." % config.sleep_for_wait)
                time.sleep(config.sleep_for_wait)
            else:
                log_error("An error occurred while getting a lock for failover execution. error=%s" % str(e))
                raise
        else:
            # Lock acquired
            log_info("Execution lock has been acquired.")
            break

    # Wait if execution lock is acquired on the other node in cluster
    lock_common.check_execution_lock_on_the_other_node(config)

    # Delete all configuration archive files so that failover_filesystem() downloads
    # the latest archive from HCP.
    print_status("Cleaning up existing archive files ...")
    archive_path_list = backup_sync_common.get_list_of_archive_files_on_HDI(config.hdi_conf_dir)
    for archive_path in archive_path_list:
        log_info("Deleting an existing archive file %s ..." % archive_path)
        Utility.remove_file(archive_path)

    # Create failover result file so that we can have at least an empty result file
    # even if there is no failover jobs.
    # Otherwise getting result at the last of failover fails with an error.
    failover_common.touch_result(config.fo_conf_dir)

    # Execute faiover for each job
    error_occurred = False
    done_names = []
    while True:
        # Get the next failover job to be processed
        name = get_next_job_name_with_lock(config.fo_conf_dir, done_names, is_error)

        # Finish if no failover jobs
        if name is None:
            print_status("=== All failover jobs are done. ===")
            break

        # Read job information
        source, fs, policy = failover_common.read_job_file(name, config.fo_conf_dir, is_error)

        # Skip the job if it doesn't have correct information
        if source is None:
            log_error("Failed to read any information from a failover job file.")
            # Delete the job file because it's not use
            delete_job_file_with_lock(name, config.fo_conf_dir, is_error)
            # Process the remaing jobs
            continue

        # Execute the failover job
        try:
            failover_filesystem.failover_filesystem(source, fs, policy, config, rehearsal)
        except Exception, e:
            # An error occurred
            error_occurred = True
            print_status_error("A failover job failed. source=%s, fs=%s, policy=%s, error=%s" % (source, fs, policy, str(e)))

            # Send an alert
            Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs, detail=str(e))

            # Mark the job as error job
            if not is_error:
                mark_job_file_as_error_with_lock(name, config.fo_conf_dir)
            else:
                # Go to next error job
                done_names.append(name)

            # Continue processing remaining jobs...
            continue

        # Finished normally. Mark the job as success job and send OK alert
        mark_job_file_as_success_with_lock(name, config.fo_conf_dir, is_error)
        Alert.send_alert(Alert.MsgCode.FAILOVER_OK, fsname=fs)

    return error_occurred

def mark_job_file_as_success_with_lock(name, config_dir, is_error=False):
    register_lock_file = lock_common.get_register_lock_file()
    with open(register_lock_file, 'w') as reg_locker:
        # Get lock for job management
        fcntl.flock(reg_locker, fcntl.LOCK_EX)
        # Mark the job file as an error job
        failover_common.mark_job_file_as_success(name, config_dir, is_error)
        # Release the lock
        fcntl.flock(reg_locker, fcntl.LOCK_UN)

def mark_job_file_as_error_with_lock(name, config_dir):
    register_lock_file = lock_common.get_register_lock_file()
    with open(register_lock_file, 'w') as reg_locker:
        # Get lock for job management
        fcntl.flock(reg_locker, fcntl.LOCK_EX)
        # Mark the job file as an error job
        failover_common.mark_job_file_as_error(name, config_dir)
        # Release the lock
        fcntl.flock(reg_locker, fcntl.LOCK_UN)

def delete_job_file_with_lock(name, config_dir, is_error=False):
    register_lock_file = lock_common.get_register_lock_file()
    with open(register_lock_file, 'w') as reg_locker:
        # Get lock for job management
        fcntl.flock(reg_locker, fcntl.LOCK_EX)
        # Delete the job file
        failover_common.delete_job_file(name, config_dir, is_error)
        # Release the lock
        fcntl.flock(reg_locker, fcntl.LOCK_UN)

def get_next_job_name_with_lock(config_dir, done_names=None, is_error=False):
    register_lock_file = lock_common.get_register_lock_file()
    with open(register_lock_file, 'w') as reg_locker:
        # Get lock for job management
        fcntl.flock(reg_locker, fcntl.LOCK_EX)
        # Get minimum number
        name = failover_common.get_next_job_name(config_dir, done_names, is_error)
        # Release the lock
        fcntl.flock(reg_locker, fcntl.LOCK_UN)
    return name

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    short_opts = "h"
    long_opts = ["id=", "dry-run", "try-out", "error", "help"]
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    rehearsal = None
    is_error = False
    for o, a in opts:
        if o == "--id":
            Utility.ID = a
        if o == "--dry-run":
            if rehearsal is not None:
                raise Exceptions.OptionError("--dry-run and --try-out cannot be specified at the same time.")
            rehearsal = failover_common.RehearsalMode(failover_common.RehearsalMode.DRYRUN)
        if o == "--try-out":
            if rehearsal is not None:
                raise Exceptions.OptionError("--dry-run and --try-out cannot be specified at the same time.")
            rehearsal = failover_common.RehearsalMode(failover_common.RehearsalMode.TRYOUT)
        if o == "--error":
            is_error = True
        if o in ("-h", "--help"):
            raise Exceptions.HelpOption()

    # Execute failover jobs
    error_occurred = execute_filesystem_failover(config, is_error=is_error, rehearsal=rehearsal)
    if error_occurred:
        log_error("An error occurred while executing a failover job.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.HelpOption:
        usage()
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, detail=str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, detail="Interrupted by a signal.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, detail="Unable to load script configuration.")
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, detail=str(e))
        sys.exit(20)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, detail=str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, detail=str(e))
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, detail=str(e))
        sys.exit(99)

    sys.exit(0)
