#!/usr/bin/python

from __future__ import with_statement
import base64
import time

import Utility
import HttpClient
try:
    from CommonLogger import log_debug, log_info, log_warning, log_error
except ImportError:
    # To make this module handy and portable
    def log_debug(msg): pass
    def log_info(msg): pass
    def log_warning(msg): print msg
    def log_error(msg): print msg

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class ConnectionError(BaseError):
    pass

class ParameterError(BaseError):
    pass

class HTTPMethod(HttpClient.HTTPMethod):
    pass

class Credential(object):
    def __init__(self, ssh_user=None, rest_user=None, ssh_key=None, rest_passMD5=None):
        self.ssh_user = ssh_user
        self.rest_user = rest_user
        self.ssh_key = ssh_key
        self.rest_passMD5 = rest_passMD5

class ConnectionSettings(object):
    def __init__(self, retry_count, retry_interval, timeout):
        self.retry_count = retry_count
        self.retry_interval = retry_interval
        self.timeout = timeout

class SSHClient(object):
    def __init__(self, host, cred, conn=None):
        self.host = host
        self.cred = cred
        self.retry_count = 0
        self.retry_interval = 0
        self.timeout = 0
        self.default_options = ["StrictHostKeyChecking no", "PasswordAuthentication no"]

        # SSH connection requires either of password or key
        if self.cred.ssh_key is None:
            log_error("SSH key was not specified for SSH connection.")
            raise ParameterError("No SSH keys for SSH connection.")

        if conn is not None:
            self.retry_count = conn.retry_count
            self.retry_interval = conn.retry_interval
            self.timeout = conn.timeout

    def set_connection_setting(self, conn):
        self.retry_count = conn.retry_count
        self.retry_interval = conn.retry_interval
        self.timeout = conn.timeout

    def test_connection(self):
        ssh_cmd = "ssh %s -i \"%s\" %s@%s '%s'" % (self._build_options(), self.cred.ssh_key, self.cred.ssh_user, self.host, 'ls')
        ret, out, err = Utility.run(ssh_cmd)
        if ret == 255:
            log_info("Test SSH connection to %s failed." % self.host)
            result = False
        else:
            log_info("Test SSH connection to %s succeeded. ret=%s, err=%s" % (self.host, ret, err))
            result = True

        return result

    def exec_cmd(self, cmd, secret_cmd=None, display_stdout=False):
        if len(cmd) == 0:
            log_error("No command was specified for SSH execution.")
            raise ParameterError("No command was specified for SSH execution.")

        ssh_cmd = "ssh -tt %s -i \"%s\" %s@%s '%s'" % (self._build_options(), self.cred.ssh_key, self.cred.ssh_user, self.host, cmd)
        if secret_cmd is not None:
            log_cmd = "ssh -tt %s -i \"%s\" %s@%s '%s'" % (self._build_options(), self.cred.ssh_key, self.cred.ssh_user, self.host, secret_cmd)
        else:
            log_cmd = ssh_cmd
        # Execute the SSH command with retry
        count = 0
        while count <= self.retry_count:
            log_debug("Command-Exec-Start: %s" % log_cmd)
            ret, out, err = Utility.run(ssh_cmd, display_stdout=display_stdout)
            log_debug("Command-Exec-End: %s, ret=%s" % (log_cmd, ret))
            if ret == 255:
                log_error("Failed to execute a command over SSH. command=\"%s\", ret=%s, out=%s, err=%s" % (log_cmd, ret, out, err))
                count += 1
                if self.retry_count != 0:
                    time.sleep(self.retry_interval)
            else:
                break
        else:
            log_error("Failed to connect the HDI over SSH. ret=%s, err=%s, host=%s" % (ret, err, self.host))
            raise ConnectionError(err, ret)

        return (ret, out, err)

    def put(self, from_file, to_path):
        log_info("Sending %s to %s:\"%s\" ..." % (from_file, self.host, to_path))
        ssh_cmd = "scp %s -i \"%s\" \"%s\" %s@%s:\"%s\"" % (self._build_options(), self.cred.ssh_key, from_file, self.cred.ssh_user, self.host, to_path)
        log_debug("Command-Exec-Start: %s" % ssh_cmd)
        ret, out, err = Utility.run(ssh_cmd)
        log_debug("Command-Exec-End: %s, ret=%s" % (ssh_cmd, ret))
        if ret != 0:
            log_error("SCP failed. ret=%s, out=%s, err=%s, from=%s, to=%s:%s" % (ret, out, err, from_file, self.host, to_path))
            raise ConnectionError(err, ret)

        return (ret, out, err)

    def get(self, from_file, to_path):
        log_info("Receiving %s from %s:\"%s\" ..." % (to_path, self.host, from_file))
        ssh_cmd = "scp %s -i \"%s\" %s@%s:\"%s\" \"%s\"" % (self._build_options(), self.cred.ssh_key, self.cred.ssh_user, self.host, from_file, to_path)
        log_debug("Command-Exec-Start: %s" % ssh_cmd)
        ret, out, err = Utility.run(ssh_cmd)
        log_debug("Command-Exec-End: %s, ret=%s" % (ssh_cmd, ret))
        if ret != 0:
            log_error("SCP failed. ret=%s, out=%s, err=%s, from=%s:%s, to=%s" % (ret, out, err, self.host, from_file, to_path))
            raise ConnectionError(err, ret)

        return (ret, out, err)

    def _build_options(self):
        options = ""
        for opt in self.default_options:
            options += "-o '%s' " % opt
        if self.timeout != 0:
            options += "-o 'ConnectTimeout %s' " % self.timeout
        return options

class MAPIClient(object):
    def __init__(self, host, cred, conn=None):
        self.host = host
        self.cred = cred
        self.retry_count = 0
        self.retry_interval = 0
        self.timeout = 0

        if cred.rest_passMD5 is None:
            log_error("Password was not specified for MAPI connection.")
            raise ParameterError("No password for MAPI connection.")

        if conn is not None:
            self.retry_count = conn.retry_count
            self.retry_interval = conn.retry_interval
            self.timeout = conn.timeout

    def set_connection_setting(self, conn):
        self.retry_count = conn.retry_count
        self.retry_interval = conn.retry_interval
        self.timeout = conn.timeout

    def exec_method(self, method, path, data=None, params=None, verbose=True, prettyprint=True):
        # params has to be a dictionary
        if method == HTTPMethod.PUT or method == HTTPMethod.POST:
            if data is None:
                raise ParameterError("No data source for PUT or POST request.")

        if method != HTTPMethod.GET:
            # verbose can be specified for GET only
            verbose = False

        if params is None:
            params = {}

        if verbose:
            params['verbose'] = "true"
        if prettyprint:
            params['prettyprint'] = None

        try:
            client = self._setup_HttpClient(params)
            if method == HTTPMethod.HEAD:
                ret = client.head(path)
            elif method == HTTPMethod.GET:
                ret = client.get(path)
            elif method == HTTPMethod.PUT:
                ret = client.put(path, data=data)
            elif method == HTTPMethod.POST:
                ret = client.post(path, data=data)
            elif method == HTTPMethod.DELETE:
                ret = client.delete(path)
        except HttpClient.ConnectionError, e:
            log_error("Failed to connect the HDI over HTTP. host=%s, path=%s, method=%s, code=%s, reason=%s." % (client.baseurl, path, HTTPMethod.tostr(method), e.code, e.reason))
            raise ConnectionError(e.reason, e.code)

        return ret

    def _setup_HttpClient(self, params):
        host = "https://%s:9090" % self.host
        client = HttpClient.HttpClient(host)

        # Setup authentication token
        auth_token = "%s:%s" % (base64.b64encode(self.cred.rest_user), self.cred.rest_passMD5)
        client.add_header(("Cookie", ("api-auth=%s" % auth_token)))

        # Setup format headers
        client.add_header(("Content-Type", "application/xml"))
        client.add_header(("Accept", "application/xml"))

        # Setup parameters
        for key in params:
            client.add_param({key: params[key]})

        # Setup retry and timeout
        client.set_retry(self.retry_count, self.retry_interval)
        client.set_timeout(self.timeout)

        return client
