#!/usr/bin/python

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "sync_settings"

import sys
import os

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import FailoverConfig
import HDIClient
import HDIOperation
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import Utility
from Utility import print_status, print_status_error

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import generate_sshkeys
import setup_common

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Load the existing pair configuration
    log_info("Getting list of pair configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Select a configuration group
    confgroup = setup_common.select_configuration_group(confgroup_list)

    # Load pair configuration
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
    pgconf = FailoverConfig.PairGroupConfig(config.fo_conf_dir, confgroup)
    stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, confgroup)

    # Setup SSHConfig
    ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

    ### Synchronize Unique SSH key to all nodes
    print_status("----- Synchronize Unique SSH key to all nodes -----")

    # Get my IP address(es)
    print_status("Getting my IP address ...")
    self_ip = setup_common.get_my_ip_address(config, confgroup)
    if self_ip is None:
        print_status_error("Could not find my IP address in the pair configuration!")
        print_status("Please check the pair configuration if this node exists in the pair configuration.")
        raise Exceptions.ConfigurationError("This node is not defined in the pair configuration.")

    # Get a list of nodes defined in the pair configuration.
    print_status("Getting a list of nodes defined in the pair configuration ...")
    node_list = pairconf.get_node_list()
    log_info("List of nodes defined in the pair configuration: %s" % node_list)

    # Generate a Unique SSH key pair if it doesn't exist yet
    if not os.path.exists(ssh_config.get_public_key()):
        # Clean up at first
        print_status("Cleaning up Unique keys ...")
        generate_sshkeys.remove_ssh_key_if_exists(ssh_config)

        # Generate Unique SSH key
        print_status("Generating a Unique key pair ...")
        generate_sshkeys.generate_keys(config.sys_conf_dir, ssh_config.get_public_key(), ssh_config.get_private_key(), confgroup=confgroup)
        try:
            print_status("Registering a new SSH key ...")
            generate_sshkeys.register_ssh_key(ssh_config.get_public_key(), confgroup=confgroup)
        except Exception, e:
            # Try rollback
            print_status_error("Something went wrong (%s). Trying rollback ..." % str(e))
            generate_sshkeys.remove_ssh_key_if_exists(ssh_config)
            raise

        # Remove default SSH key
        try:
            print_status("Removing default SSH key ...")
            generate_sshkeys.remove_ssh_key_if_exists(ssh_config, is_default=True)
        except Exception, e:
            # Try rollback
            print_status_error("Something went wrong (%s). Trying rollback ..." % str(e))
            generate_sshkeys.register_ssh_key(ssh_config.get_public_key(is_default=True), is_default=True)
            raise

    # Mark myself as 'done'
    pairconf.done_ssh_setup(self_ip)
    pairconf.save()

    # Setup Unique SSH key on the nodes where Unique SSH key is not setup yet.
    for node in node_list:
        # Is this myself?
        if node == self_ip:
            # Already done
            continue

        # Prepare transfer
        cred_defkey = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(is_default=True), None)
        conn = HDIOperation.Connection(node, cred_defkey, config.retry_count, config.retry_interval, config.timeout)

        # Test connection to the node with Unique SSH key
        # It is possible that we can connect the node with Unique SSH key,
        # if the node was deleted from pair configuration temporarily
        cred_uniqkey = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(), None)
        sshclient = HDIClient.SSHClient(node, cred_uniqkey)
        if sshclient.test_connection():
            print_status("Unique SSH key is already set up on %s." % node)
            # Mark as 'done'
            pairconf.done_ssh_setup(node)
            pairconf.save()
            continue

        # Remove the existing Unique keys on the node
        print_status("Removing SSH key on %s ..." % node)
        cmd = "sudo maintexec %s --remove %s" % (os.path.join(BASEDIR, "int/generate_sshkeys"), ssh_config.confgroup)
        op = HDIOperation.CommandOperation(cmd)
        try:
            ret, out, err = op.execute(conn)
        except HDIOperation.ConnectionError, e:
            print_status_error("Failed to communicate with %s. error=%s" % (conn.host, str(e)))
            raise Exceptions.ConnectionError(e.reason, e.code)
        if ret != 0:
            print_status_error("Failed to remove Unique SSH keys on %s. ret=%s, out=%s, err=%s" % (conn.host, ret, out, err))
            raise Exceptions.OperationError("SSH key removal failure. host=%s, error=%s" % (conn.host, err), ret)

        # Transfer the public key
        print_status("Sending SSH keys to %s ..." % node)
        op = HDIOperation.TransferOperation(ssh_config.get_public_key(), ssh_config.get_public_key())
        try:
            op.send(conn)
        except HDIOperation.ConnectionError, e:
            print_status_error("Failed to send public SSH key to %s. error=%s" % (conn.host, str(e)))
            raise Exceptions.ConnectionError(e.reason, e.code)

        # Transfer the private key
        op = HDIOperation.TransferOperation(ssh_config.get_private_key(), ssh_config.get_private_key())
        try:
            op.send(conn)
        except HDIOperation.ConnectionError, e:
            log_error("Failed to send private SSH key to %s. error=%s" % (conn.host, str(e)))
            raise Exceptions.ConnectionError(e.reason, e.code)

        # Register the public key
        print_status("Registering SSH key on %s ..." % node)
        cmd = "sudo maintexec %s --register %s" % (os.path.join(BASEDIR, "int/generate_sshkeys"), ssh_config.confgroup)
        op = HDIOperation.CommandOperation(cmd)
        try:
            ret, out, err = op.execute(conn)
        except HDIOperation.ConnectionError, e:
            log_error("Failed to communicate with %s. error=%s" % (conn.host, str(e)))
            raise Exceptions.ConnectionError(e.reason, e.code)

        if ret != 0:
            log_error("Failed to register the SSH public key on %s. ret=%s, out=%s, err=%s" % (conn.host, ret, out, err))
            raise Exceptions.OperationError("SSH key registration failure. host=%s, error=%s" % (conn.host, err), ret)

        # Remove default SSH key
        print_status("Removing default SSH key on %s ..." % node)
        cmd = "sudo maintexec %s --remove" % os.path.join(BASEDIR, "int/install_default_key")
        op = HDIOperation.CommandOperation(cmd)
        try:
            ret, out, err = op.execute(conn)
        except HDIOperation.ConnectionError, e:
            log_error("Failed to communicate with %s. error=%s" % (conn.host, str(e)))
            raise Exceptions.ConnectionError(e.reason, e.code)

        if ret != 0:
            log_error("Failed to remove the default SSH key on %s. ret=%s, error=%s" % (conn.host, ret, err))
            raise Exceptions.OperationError("SSH key remove failure. host=%s, error=%s" % (conn.host, err), ret)

        # Mark as 'done'
        pairconf.done_ssh_setup(node)
        pairconf.save()

    ### Synchronize pair/pair group/storage policy configurations to all nodes
    print_status("----- Synchronizing failover configuration with all nodes -----")

    for node in node_list:
        if node == self_ip:
            # Exclude myself
            continue

        print_status("Transferring failover configuration files to %s ..." % node)
        # Transfer preparation
        cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(node), None)
        conn = HDIOperation.Connection(node, cred, config.retry_count, config.retry_interval, config.timeout)

        # Transfer pair, pair group and storage policy configuration files
        pair_op = HDIOperation.TransferOperation(pairconf.config_file, pairconf.config_file)
        pg_op = HDIOperation.TransferOperation(pgconf.config_file, pgconf.config_file)
        st_op = HDIOperation.TransferOperation(stconf.config_file, stconf.config_file)
        try:
            pair_op.send(conn)
            st_op.send(conn)
            # Pair group configuration may not exist if not configured
            if os.path.exists(pgconf.config_file):
                pg_op.send(conn)
        except HDIOperation.ConnectionError, e:
            log_error("Failed to send configuration files to %s. error=%s" % (conn.host, str(e)))
            raise Exceptions.ConnectionError(e.reason, e.code)

    print_status("Completed.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
